/*********************************************************************************
**
**   $Id: DataForm.cpp,v 1.16 2004/06/28 11:58:28 croftj Exp $
**   Copyright (c) 2004 Joe Croft <joe@croftj.net>
**   
**   This file is part of ah2.
**
**   ah2 is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 2 of the License, or
**   (at your option) any later version.
**
**   Foobar is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with Foobar; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
*********************************************************************************/

#include <qfont.h>
#include <qfontmetrics.h>
#include <qlabel.h>
#include <qlistbox.h>
#include <qmessagebox.h>
#include <qregexp.h>
#include <qsqlcursor.h>
#include <qsqldatabase.h>
#include <qsqldriver.h>
#include <qsqlfield.h>
#include <qsqlform.h>
#include <qsqlpropertymap.h>
#include <qsqlrecord.h>
#include <stdio.h>
#include "Data/DataForm.h"

MoneyEdit::MoneyEdit(QWidget *parent, const char *name) : QLineEdit (parent, name) {}   
MoneyEdit::MoneyEdit(const QString &contents, QWidget *parent, const char *name) : QLineEdit(contents, parent, name) {}

double MoneyEdit::text() const
{
   QString s = QLineEdit::text();
   printf("MoneyEdit::text(): started with |%s|\n", (const char*)s);
   s.replace(QRegExp("[\\$,]"), "");
   printf("MoneyEdit::text(): returned |%s|\n", (const char*)s);
   return(s.toDouble());
}

void MoneyEdit::setText(const double f) 
{
   QLineEdit::setText(formatCurrency(f));
}

QString MoneyEdit::formatCurrency(double f)
{
   QString s, s1, s2;

   s.sprintf("%15.2f", f);
   QRegExp re1("([0-9])([0-9]{3}\\.)");
   re1.setMinimal(true);
   re1.search(s);
   s1 = re1.cap(1);
   s2 = re1.cap(2);
   printf("MoneyEdit::setText(): s1 = |%s|, s2 = |%s|\n", (const char*)s1, (const char*)s2);
   s.replace(re1, s1 + "," + s2);

   QRegExp re2("([0-9])([0-9]{3},)");
   re2.setMinimal(true);

   re2.search(s);
   s1 = re2.cap(1);
   s2 = re2.cap(2);
   s.replace(re2, s1 + "," + s2);

   re2.search(s);
   s1 = re2.cap(1);
   s2 = re2.cap(2);
   s.replace(re2, s1 + "," + s2);

   re2.search(s);
   s1 = re2.cap(1);
   s2 = re2.cap(2);
   s.replace(re2, s1 + "," + s2);

   s.replace(QRegExp("^\\s*"), "");
   s = "$" + s;
   return(s);
}

QString StringSelect::text() const
{
   QString s = currentText();
   if ( s == "<NULL>") 
   {
      s = "";
   }
   return(s);
}

void StringSelect::setText(QString s)
{
   int i = 0;
   if ( s == "" ) 
   {
      s = "<NULL>";
   }
   QListBoxItem *li = listBox()->findItem(s);
   if ( li != NULL ) 
   {
      i = listBox()->index(li);
      listBox()->setSelected(li, true);
      setCurrentItem(i);
      setCurrentText(s);
   }
   else 
   {
      listBox()->insertItem(s, 0);
      setCurrentItem(0);
      setCurrentText(s);
   }
}

void StringSelect::initialize(QString s)
{
   printf("StringSelect::initialize(): s = |%s|\n", (const char*)s);
   QStringList namesList = QStringList::split( ",", s );
   clear();
   insertStringList(namesList);
}



QString YesNoSelect::text() const
{
   QString rv;
   if ( currentText() == "Yes" ) 
      rv = "Y";
   else
      rv = "N";
   return(rv);
}

void YesNoSelect::setText(QString t)
{
   QString s;
   if ( t == "Y" ) 
      s = "Yes";
   else
      s = "No";
   QListBoxItem *li = listBox()->findItem(s);
   int i = listBox()->index(li);
   listBox()->setSelected(li, true);
   setCurrentItem(i);
   setCurrentText(s);
}

void YesNoSelect::initialize()
{
   QStringList typeList;
   typeList += "";
   typeList += "Yes";
   typeList += "No";
   clear();
   insertStringList(typeList);
}

DataForm::DataForm(QSqlDatabase *_db, const QString table, int rows, int cols, QWidget *pParent, const char *pName, WFlags f ) :
          QFrame(pParent, pName, f)
{
   printf("DataForm::DataForm(): Enter, table = |%s|\n", (const char*)table);
   fflush(stdout);
   db = _db;
   dbTable = table;
   gl = new QGridLayout(this, rows, cols * 2, 5, 5);
   gl->activate();
   sqlForm     = NULL;
   sqlRecord   = NULL;
   sqlCursor   = NULL;
   ident = "";
   insertMode  = false;
   inTransaction = false;
   setFocusPolicy(QWidget::StrongFocus);
   printf("DataForm::DataForm(): Exit\n");
   fflush(stdout);
}

void DataForm::setFields(const DataFields *fields) 
{
   QWidget *wdt, *lastwdt = NULL;
   QSizePolicy labelPolicy(QSizePolicy::Minimum, QSizePolicy::Preferred);

   printf("DataForm::setFields(): Enter\n");
   fflush(stdout);
   QSqlDriver *dbdrv = db->driver();
   printf("DataForm::setFields(): Getting Record Info from (%ld:%s)\n", (long)dbdrv, (const char *)dbTable);
   fflush(stdout);
   QSqlRecordInfo recInfo = dbdrv->recordInfo(dbTable);

   setFocusPolicy(QWidget::StrongFocus);

   printf("DataForm::setFields(): Iterating field info\n");
   fflush(stdout);

   //
   // Iterate through all of the field definitions and 
   // place them into the grid.
   //
   while (fields->dataName != NULL) 
   {
      if ( fields->widget == NULL ) 
      {
         printf("DataForm::setFields(): New edit field\n");
         fflush(stdout);
         wdt = new QLineEdit(this);
         if ( fields->ro ) 
            ((QLineEdit*)wdt)->setReadOnly(true);
      }
      else 
      {
         printf("DataForm::setFields(): Existing Widget\n");
         fflush(stdout);
         wdt = fields->widget;
      }
      if ( lastwdt != NULL ) 
         setTabOrder(lastwdt, wdt);
      lastwdt = wdt;

      QFont f = wdt->font();
      QFontMetrics fm(f);
      int fwidth = fm.maxWidth();
      printf("DataForm::setFields():Element |%s|\n", (const char*)fields->dataName);
      fflush(stdout);
      if ( fields->minWidth > 0 ) 
      {
         printf("DataForm::setFields():Setting min width %d\n", fwidth * fields->minWidth);
         fflush(stdout);
         wdt->setMinimumWidth(fwidth * fields->minWidth);
      }
      if ( fields->maxWidth > 0 ) 
      {
         printf("DataForm::setFields():Setting max width %d\n", fwidth * fields->maxWidth);
         fflush(stdout);
         wdt->setMaximumWidth(fwidth * fields->maxWidth);
      }
      QSqlFieldInfo fi = recInfo.find(fields->dataName);
      printf("DataForm::setFields():Setting max length %d\n", fi.length());
      QVariant v = fi.type();
      printf("DataForm::setFields():Field type: |%s|\n", (const char*)v.toString());
      printf("DataForm::setFields():Widget Class: |%s|\n", (const char*)wdt->className());
      printf("DataForm::setFields():Field type: |%d|\n", fi.typeID());
      fflush(stdout);
//      if ( fi.typeID() == 1043 ) 
//         wdt->setMaxLength(fi.length());
      formFields[fields->dataName] = wdt;

      if ( fields->propName != NULL ) 
         widgetProperties[fields->dataName] = fields->propName;

      printf("DataForm::setFields():Default Text: |%s|\n", (const char*)fields->defvalue);
      fflush(stdout);
      fieldDefaults[fields->dataName] = fields->defvalue;

      
      printf("DataForm::setFields(): rowSpan = %d\n", fields->rowSpan);
      printf("DataForm::setFields(): colSpan = %d\n", fields->colSpan);
      if ( fields->rowSpan <= 0) 
      {
         printf("DataForm::setFields(): Adding label- row = %d, col = %d\n", fields->row, fields->col * 2);
         QLabel *lbl = new QLabel(fields->label, this);
         lbl->setSizePolicy(labelPolicy);
         lbl->setAlignment(Qt::AlignTop|Qt::AlignRight);
         gl->addWidget(lbl, fields->row, fields->col * 2);
      }
      else 
      {
         int rows = (fields->rowSpan == 1) ? 0 : fields->rowSpan;
         printf("DataForm::setFields(): rows = %d\n", rows);
         printf("DataForm::setFields(): Adding multi row label- beg row = %d, end row = %d beg col = %d, end col = %d\n", 
                                       fields->row, fields->row + rows, (fields->col * 2), (fields->col * 2));
         QLabel *lbl = new QLabel(fields->label, this);
         lbl->setAlignment(Qt::AlignTop|Qt::AlignRight);
         gl->addMultiCellWidget(lbl, fields->row, fields->row + rows, (fields->col * 2), (fields->col * 2));
      }
      if ( fields->colSpan > 1 || fields->rowSpan > 1) 
      {
         int rows = (fields->rowSpan <= 1) ? 1 : fields->rowSpan;
         printf("DataForm::setFields(): rows = %d\n", rows);
         int cols = (fields->colSpan <= 1) ? 1 : fields->colSpan;
         printf("DataForm::setFields(): cols = %d\n", cols);
         printf("DataForm::setFields(): Adding multicell widget- beg row = %d, end row = %d beg col = %d, end col = %d\n", 
                                        fields->row, fields->row + rows - 1, (fields->col * 2) + 1, ((fields->col * 2) + cols * 2) - 1);
         gl->addMultiCellWidget(wdt, fields->row, fields->row + rows - 1, (fields->col * 2) + 1, ((fields->col * 2) + cols * 2) - 1);
      }
      else
      {
         gl->addWidget(wdt, fields->row, (fields->col * 2) + 1);
      }
      fields++;
   }
   printf("DataForm::setFields(): Exit\n");
   fflush(stdout);
}

void DataForm::refresh(QString item)
{
   printf("DataForm::refresh(): Enter insertMode = %d\n", insertMode);
   fflush(stdout);
   if ( ! insertMode ) 
   {
      ident = item;
      strIdent = "";
      strIdent += "ident='" + item + "'";
      printf("DataForm::refresh(): strIdent = %s\n", (const char *)strIdent);
      fflush(stdout);

      if ( sqlForm   != NULL ) delete sqlForm;

      sqlCursor = new QSqlCursor(dbTable);
      sqlCursor->select(strIdent);
      sqlCursor->first();
      sqlRecord = sqlCursor->primeUpdate();

      printf("DataForm::refresh(): Found item |%s|\n", (const char *)sqlRecord->value("ident").toString());
      fflush(stdout);

      sqlForm = new QSqlForm(this);
      sqlForm->setRecord(sqlRecord);

      //
      // Install property maps for any special widgets
      //
      QSqlPropertyMap *pmap = new QSqlPropertyMap;
      QMap<QString, QString>::Iterator it;
      for (it = widgetProperties.begin(); it != widgetProperties.end(); ++it) 
      {
         pmap->insert(formFields[it.key()]->className(), it.data());
      }
      sqlForm->installPropertyMap(pmap);

      //
      // Insert the regular widgets into the form
      //
      QMap<QString, QWidget*>::Iterator editIt;
      for (editIt = formFields.begin(); editIt != formFields.end(); ++editIt) 
      {
         printf("Adding field |%s| to form\n", (const char*)editIt.key());
         fflush(stdout);
         sqlForm->insert(editIt.data(), editIt.key());
      }

      //
      // Now add in the special widgets to the form
      //
      QMap<QString, QWidget*>::Iterator widgetIt;
      for (widgetIt = formWidgets.begin(); widgetIt != formWidgets.end(); ++widgetIt) 
      {
         printf("Adding field |%s| to form\n", (const char*)widgetIt.key());
         fflush(stdout);
         sqlForm->insert(widgetIt.data(), widgetIt.key());
      }
      sqlForm->readFields();
   }
   printf("DataForm::refresh():Exit\n");
}

void DataForm::refresh(QSqlRecord *r)
{
   printf("DataForm::refresh():Enter (qsqlrecord)\n");
   refresh(r->value("ident").toString());
   printf("DataForm::refresh():Exit (qsqlrecord)\n");
}

void DataForm::updateRecord()
{
   printf("DataForm::updateRecord():Enter()\n");
   fflush(stdout);

   if ( sqlForm != NULL ) 
   {
      sqlForm->writeFields();
      int x = sqlCursor->update();
      printf("DataForm::updateRecord():Updated %d records, db driver error = |%s|\n", x, (const char*)sqlCursor->lastError().driverText());
      printf("DataForm::updateRecord():Updated %d records, database  error = |%s|\n", x, (const char*)sqlCursor->lastError().databaseText());
      fflush(stdout);
      sqlCursor->select(strIdent);
      sqlCursor->first();
      sqlForm->readFields();
      emit(updated());
      insertMode = false;
   }
}

void DataForm::deleteRecord()
{
   printf("DataForm::deleteRecord():Enter()\n");
   fflush(stdout);

   sqlRecord = sqlCursor->primeDelete();
   int x = sqlCursor->del();
   printf("DataForm::deleteRecord():Updated %d records, db driver error = |%s|\n", x, (const char*)sqlCursor->lastError().driverText());
   printf("DataForm::deleteRecord():Updated %d records, database  error = |%s|\n", x, (const char*)sqlCursor->lastError().databaseText());
   fflush(stdout);
   insertMode = false;
   emit(updated());
}



void DataForm::insertRecord()
{
   QMap<QString, QString> fields;
   QString idx, sBreed, sql;

   if ( ! inTransaction ) 
   {
      QMessageBox::warning(NULL, "Error!", "You must begin a transaction before inserting reecords", "Okay");
      return;
   }

   if ( insertMode == true ) 
   {
      updateRecord();
   }

   QString seq = dbTable + "_ident_seq";
   sql = "select nextval('" + seq + "')";
   printf("DataForm::insertRecord():sql = '%s'\n", (const char*)sql);
   fflush(stdout);
   QSqlQuery q1(sql);
   if ( q1.isActive() )
   {
      if ( q1.next() ) 
      {
         idx = q1.value(0).toString();
         printf("DataForm::insertRecord():new idx = '%s'\n", (const char*)idx);
         fflush(stdout);
         sql = "insert into " + dbTable + " (ident) values ('" + idx + "')";
         QSqlQuery q2(sql);
         if ( !q2.isActive() ) 
            QMessageBox::warning(NULL, "Error!", q2.lastError().driverText() + q2.lastError().databaseText(), "Okay");
         else 
         {
            refresh(idx);
            QMap<QString, QString>::Iterator it;
            for (it = fieldDefaults.begin(); it != fieldDefaults.end(); ++it) 
            {
               printf("setting Text: |%s|\n", (const char*)it.data());
//               formFields[it.key()]->setText(it.data());
            }
         }
      }
   }
   insertMode = true;
}

void DataForm::beginTransaction()
{
   db->transaction();
   inTransaction = true;
}

void DataForm::commitTransaction()
{
   db->commit();
   insertMode = false;
   inTransaction = false;
   emit(updated());
}

void DataForm::rollbackTransaction()
{
   if ( inTransaction ) 
   {
      db->rollback();
      insertMode = false;
      inTransaction = false;
      emit(updated());
   }
   else 
   {
      sqlCursor->select(strIdent);
      sqlCursor->first();
      sqlForm->readFields();
   }
}

QString DataForm::getLastIdent()
{
   printf("DataForm::getLastIdent():strIdent = '%s'\n", (const char*)ident);
   return(ident);
}
